# Usage: python umapper.py <dataset_id> <model> <neighbors> <min_dist>
# Example: python umapper.py dadabase-curated BAAI_bge-small-en-v1.5 50 0.075
import os
import re
import sys
import json
import argparse

from latentscope.util import get_data_dir

def main():
    parser = argparse.ArgumentParser(description='Generate SAE features from embeddings for a dataset')
    parser.add_argument('dataset_id', type=str, help='Dataset name (directory name in data/)')
    parser.add_argument('embedding_id', type=str, help='Name of embedding to use')
    parser.add_argument("model_id", type=str, nargs="?", help="HF id of model to use", default="enjalot/sae-nomic-text-v1.5-FineWeb-edu-10BT")
    parser.add_argument('k_expansion', type=str, nargs="?", help='Output file', default="64_32")
    parser.add_argument('device', type=str, nargs="?", help='Device to use')

    # Parse arguments
    args = parser.parse_args()
    saer(args.dataset_id, args.embedding_id, args.model_id, args.k_expansion, args.device)


def saer(dataset_id, embedding_id, model_id, k_expansion, device):
    DATA_DIR = get_data_dir()
    # read in the embeddings 

    sae_dir = os.path.join(DATA_DIR, dataset_id, "saes")
    if not os.path.exists(sae_dir):
        os.makedirs(sae_dir)

    # determine the index of the last sae run by looking in the dataset directory
    # for files named sae-<number>.json
    sae_files = [f for f in os.listdir(sae_dir) if re.match(r"sae-\d+\.json", f)]
    if len(sae_files) > 0:
        last_sae = sorted(sae_files)[-1]
        last_sae_number = int(last_sae.split("-")[1].split(".")[0])
        next_sae_number = last_sae_number + 1
    else:
        next_sae_number = 1

    # make the sae name from the number, zero padded to 3 digits
    sae_id = f"sae-{next_sae_number:03d}"
    print("RUNNING:", sae_id)

    from latentsae.sae import Sae
    import h5py
    import numpy as np
    import pandas as pd
    import torch

    print("loading embeddings")
    embedding_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{embedding_id}.h5")
    with h5py.File(embedding_path, 'r') as f:
        dataset = f["embeddings"]
        embeddings = np.array(dataset)

    if device == "mps" or torch.backends.mps.is_available():
        device = torch.device("mps")
    elif device == "cuda" or torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    all_embeddings = torch.from_numpy(embeddings).float().to(device)
    model = Sae.load_from_hub(model_id, k_expansion, device)

    print("Encoding embeddings with SAE")
    batch_size = 128  # Define your batch size
    # all_features = []
    all_acts = []
    all_indices = []

    from tqdm import tqdm  # Make sure to import tqdm at the top of your file

    for i in tqdm(range(0, len(all_embeddings), batch_size), desc="Encoding batches"):
        batch = all_embeddings[i:i + batch_size]
        features = model.encode(batch)
        all_acts.append(features.top_acts)
        all_indices.append(features.top_indices)

    all_acts = torch.cat(all_acts, dim=0).detach().cpu().numpy()
    all_indices = torch.cat(all_indices, dim=0).detach().cpu().numpy()

    # all_features = model.encode(all_embeddings)

    print("encoding completed")
    # all_acts = all_features.top_acts.detach().cpu().numpy()
    # all_indices = all_features.top_indices.detach().cpu().numpy()

    print("saving to disk")
    # save the acts and indices to the sae directory
    with h5py.File(os.path.join(sae_dir, f"{sae_id}.h5"), 'w') as f:
        f.create_dataset("top_acts", data=all_acts)
        f.create_dataset("top_indices", data=all_indices)

    print("calculating summary statistics")

    import scipy
    print("ALL ACTS SHAPE", all_acts.shape)
    print("ALL INDS SHAPE", all_indices.shape)
    matrix = scipy.sparse.lil_matrix((all_acts.shape[0], model.num_latents), dtype=np.float32)
    # matrix.rows = all_indices.tolist()
    # matrix.data = all_acts.tolist()
    for i in range(all_acts.shape[0]):
        matrix.rows[i] = all_indices[i].tolist()
        matrix.data[i] = all_acts[i].tolist()
    print("MATRIX", matrix.shape)

    csr = matrix.tocsr()
    max_activations = csr.max(axis=0).toarray().flatten()

    print("MAX ACTIVATIONS", max_activations.shape)
    print("MAX ACTIVATIONS", max_activations)

    # TODO a faster way to calculate these? probably a scikit sparse matrix calculation or something
    # for feature_idx in tqdm(range(model.num_latents), desc="Calculating max activations"):
    #     feature_mask = all_indices == feature_idx
    #     if feature_mask.any():
    #         feature_activations = np.where(feature_mask, all_acts, 0.0)
    #         max_activation = np.max(feature_activations)
    #         max_activations[feature_idx] = max_activation

    dead_features = np.where(max_activations <= 0)
    alive_features = np.where(max_activations > 0)
    # Count the number of dead features
    num_dead_features = len(dead_features[0])
    print(f"Number of dead features: {num_dead_features}")

    # create the metadata json
    with open(os.path.join(sae_dir, f"{sae_id}.json"), 'w') as f:
        json.dump({
            "id": sae_id,
            "model_id": model_id,
            "k_expansion": k_expansion,
            "embedding_id": embedding_id,
            "dataset_id": dataset_id,
            "max_activations": [round(act, 5) for act in max_activations.tolist()],
            "num_features": model.num_latents,
            "dead_features": num_dead_features,
            "alive_features": len(alive_features[0])
        }, f)

    print("done with", sae_id)


if __name__ == "__main__":
    main()



from .base import EmbedModelProvider, ChatModelProvider

class TransformersEmbedProvider(EmbedModelProvider):
    def __init__(self, name, params):
        super().__init__(name, params)
        import torch
        self.torch = torch
        self.device = torch.device("cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu")
        
    def load_model(self):
        # from transformers import AutoTokenizer, AutoModel
        from sentence_transformers import SentenceTransformer
        self.model = SentenceTransformer(self.name, trust_remote_code=True, device=self.device)#, backend="onnx")
        self.tokenizer = self.model.tokenizer

    def embed(self, inputs, dimensions=None):
        embeddings = self.model.encode(inputs, convert_to_tensor=True)
        # Support Matroyshka embeddings
        if dimensions is not None and dimensions > 0:
            embeddings = self.torch.nn.functional.layer_norm(embeddings, normalized_shape=(embeddings.shape[1],))
            embeddings = embeddings[:, :dimensions]

        # Normalize embeddings
        normalized_embeddings = self.torch.nn.functional.normalize(embeddings, p=2, dim=1)
        return normalized_embeddings.tolist()


class TransformersChatProvider(ChatModelProvider):
    def __init__(self, name, params):
        super().__init__(name, params)
        import torch
        from transformers import pipeline
        import outlines
        self.torch = torch
        self.pipeline = pipeline
        self.outlines = outlines
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def load_model(self):
        # TODO: support bfloat16 for non mac environments
        from transformers import AutoTokenizer
        self.model = self.outlines.models.transformers(self.name)
        self.generator = self.outlines.generate.text(self.model)
        # TODO: this had an error. but also it would exclude foreign characters
        # self.generator = self.outlines.generate.regex(
        #     self.model
        #     , r"[a-zA-Z0-9 ]+"
        #     )
        self.encoder = AutoTokenizer.from_pretrained(self.name, trust_remote_code=True)
        # self.pipe = self.pipeline("text-generation", model=self.name, torch_dtype=self.torch.float16, device=self.device, trust_remote_code=True)

    def summarize(self, items, context=""):
        # from tqdm import tqdm
        from .prompts import summarize, summarize_system_prompt
        prompt = summarize(items, context)
        # tqdm.write(f"Summarizing {prompt}")
        # Apply chat template to format prompt for model
        formatted_prompt = self.encoder.apply_chat_template([
            # {"role": "system", "content": summarize_system_prompt},
            {"role": "user", "content": prompt}], 
            tokenize=False,
            add_generation_prompt=True)
        return self.generator(formatted_prompt)



import os
import sys
import json
import numpy as np
import pandas as pd
from datetime import datetime
from scipy.spatial import ConvexHull
from flask import Blueprint, jsonify, request

# Create a Blueprint
bulk_bp = Blueprint('bulk_bp', __name__)
bulk_write_bp = Blueprint('bulk_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

def write_transaction(dataset_id, scope_id, action, payload):
    transactions_file_path = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + "-transactions.json")
    if not os.path.exists(transactions_file_path):
        with open(transactions_file_path, 'w') as f:
            json.dump([], f)
    with open(transactions_file_path, 'r') as f:
        transactions = json.load(f)
    new_transaction = {
        "action": action,
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "payload": payload
    }
    transactions.append(new_transaction)
    with open(transactions_file_path, 'w') as f:
        json.dump(transactions, f, indent=2)

# Change the cluster of rows
@bulk_write_bp.route('/change-cluster', methods=['POST'])
def change_cluster():
  data = request.get_json()
  print(data)
  dataset_id = data["dataset_id"]
  scope_id = data["scope_id"]
  row_ids = data["row_ids"]
  new_cluster = int(data["new_cluster"])

  # read in the scope parquet
  scope_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".parquet")
  df = pd.read_parquet(scope_file)

  # read the scope metadata for the cluster_label_map
  scope_meta_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".json")
  with open(scope_meta_file) as f:
    scope_meta = json.load(f)


  clusters = scope_meta["cluster_labels_lookup"]
  new_label = clusters[new_cluster]
  # change the cluster of the rows
  df.loc[df['ls_index'].isin(row_ids), "cluster"] = new_cluster
  df.loc[df['ls_index'].isin(row_ids), "label"] = new_label["label"]

  # write the parquet
  df.to_parquet(scope_file)
  update_combined(df, dataset_id, scope_id)

  # recalculate the hulls
  for c in clusters:
      indices = df[df['cluster'] == c["cluster"]]["ls_index"].tolist()
      label_points = df[df['cluster'] == c["cluster"]][['x', 'y']].values
      if(len(label_points) > 0):
        hull = ConvexHull(label_points)
        hull_list = [indices[s] for s in hull.vertices.tolist()]
        c["hull"] = hull_list
      else:
        c["hull"] = []

  #write the scope meta to file
  with open(scope_meta_file, "w") as f:
    json.dump(scope_meta, f, indent=2)

  write_transaction(dataset_id, scope_id, "change_cluster", {
    "row_ids": row_ids,
    "new_cluster": new_cluster
  })

  return jsonify({"success": True})

# change cluster name
# Change the cluster of rows
@bulk_write_bp.route('/change-cluster-name', methods=['GET'])
def change_cluster_name():
  dataset_id = request.args["dataset_id"]
  scope_id = request.args["scope_id"]
  cluster = int(request.args["cluster"])
  new_label = request.args["new_label"]

  # read in the scope parquet
  scope_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".parquet")
  df = pd.read_parquet(scope_file)

  # read the scope metadata for the cluster_label_map
  scope_meta_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".json")
  with open(scope_meta_file) as f:
    scope_meta = json.load(f)

  clusters = scope_meta["cluster_labels_lookup"]
  updated = clusters[cluster]
  updated["label"] = new_label

  # update the label column in the parquet
  df[df['cluster'] == cluster]['label'] = new_label
  df.to_parquet(scope_file)

  update_combined(df, dataset_id, scope_id)

  #write the scope meta to file
  with open(scope_meta_file, "w") as f:
    json.dump(scope_meta, f, indent=2)

  write_transaction(dataset_id, scope_id, "change_cluster_name", {
    "cluster": cluster,
    "new_label": new_label
  })
  return jsonify({"success": True})


# Delete rows
@bulk_write_bp.route('/delete-rows', methods=['POST'])
def delete_rows():
  data = request.get_json()
  print(data)
  dataset_id = data["dataset_id"]
  scope_id = data["scope_id"]
  row_ids = data["row_ids"]
  # read in the scope parquet
  scope_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".parquet")
  df = pd.read_parquet(scope_file)
  # df = df[~df['ls_index'].isin(row_ids)]
  # df.reset_index(drop=True, inplace=True)
  if 'deleted' in df.columns:
      df.loc[df['ls_index'].isin(row_ids), "deleted"] = True

  df.to_parquet(scope_file)

  update_combined(df, dataset_id, scope_id)

  # recalculate hulls
  scope_meta_file = os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + ".json")
  with open(scope_meta_file) as f:
    scope_meta = json.load(f)
  clusters = scope_meta["cluster_labels_lookup"]
  for c in clusters:
    indices = df[df['cluster'] == c["cluster"]]["ls_index"].tolist()
    label_points = df[df['cluster'] == c["cluster"]][['x', 'y']].values
    try:
      hull = ConvexHull(label_points)
      hull_list = [indices[s] for s in hull.vertices.tolist()]
      c["hull"] = hull_list
    except:
      c["hull"] = []

  scope_meta["rows"] = len(df)
  # write the scope meta to file
  with open(scope_meta_file, "w") as f:
    json.dump(scope_meta, f, indent=2)

  write_transaction(dataset_id, scope_id, "delete_rows", {
    "row_ids": row_ids
  })
  return jsonify({"success": True})

def update_combined(df, dataset_id, scope_id):

  
  input_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "input.parquet"))
  input_df.reset_index(inplace=True)
  input_df = input_df[input_df['index'].isin(df['ls_index'])]
  combined_df = input_df.join(df.set_index('ls_index'), on='index', rsuffix='_ls')
  combined_df.to_parquet(os.path.join(DATA_DIR, dataset_id, "scopes", scope_id + "-input.parquet"))



import os
import re
import csv
import json
import uuid
from importlib.resources import files
from flask import Blueprint, jsonify, request

# Create a Blueprint
models_bp = Blueprint('models_bp', __name__)
models_write_bp = Blueprint('models_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

@models_bp.route('/embedding_models', methods=['GET'])
def get_embedding_models():
    embedding_path = files('latentscope.models').joinpath('embedding_models.json')
    with embedding_path.open('r', encoding='utf-8') as file:
        models = json.load(file)
    return jsonify(models)

@models_bp.route('/chat_models', methods=['GET'])
def get_chat_models():
    chat_path = files('latentscope.models').joinpath('chat_models.json')
    with chat_path.open('r', encoding='utf-8') as file:
        models = json.load(file)
    return jsonify(models)

@models_bp.route('/embedding_models/recent', methods=['GET'])
def get_recent_embedding_models():
    return get_recent_models("embedding")

@models_bp.route('/chat_models/recent', methods=['GET'])
def get_recent_chat_models():
    return get_recent_models("chat")

def get_recent_models(model_type="embedding"):
    recent_models_path = os.path.join(DATA_DIR, f"{model_type}_model_history.csv")
    if not os.path.exists(recent_models_path):
        return jsonify([])

    with open(recent_models_path, 'r', encoding='utf-8') as file:
        reader = csv.reader(file)
        recent_models = []
        for row in reader:
            recent_models.append({
                "timestamp": row[0],
                "id": row[1],
                "group": "recent",
                "provider": row[1].split("-")[0],
                "name": "-".join(row[1].split("-")[1:]).replace("___", "/")
            })
    recent_models.sort(key=lambda x: x["timestamp"], reverse=True)
    # Deduplicate models with the same id
    seen_ids = set()
    unique_recent_models = []
    for model in recent_models:
        if model["id"] not in seen_ids:
            unique_recent_models.append(model)
            seen_ids.add(model["id"])
    recent_models = unique_recent_models[:5]
    return jsonify(recent_models)

@models_bp.route('/custom-models', methods=['GET'])
def get_custom_models():
    custom_models_path = os.path.join(DATA_DIR, "custom_models.json")
    if not os.path.exists(custom_models_path):
        return jsonify([])
    with open(custom_models_path, 'r', encoding='utf-8') as file:
        custom_models = json.load(file)
    return jsonify(custom_models)

@models_write_bp.route('/custom-models', methods=['POST'])
def add_custom_model():
    data = request.json
    custom_models_path = os.path.join(DATA_DIR, "custom_models.json")
    
    # Read existing models
    existing_models = []
    if os.path.exists(custom_models_path):
        with open(custom_models_path, 'r', encoding='utf-8') as file:
            existing_models = json.load(file)
    
    # Add new model
    data["id"] = "custom-" + data["name"]
    existing_models.append(data)
    
    # Write updated models
    with open(custom_models_path, 'w', encoding='utf-8') as file:
        json.dump(existing_models, file)
        
    return jsonify(existing_models)

@models_write_bp.route('/custom-models/<model_id>', methods=['DELETE'])
def delete_custom_model(model_id):
    custom_models_path = os.path.join(DATA_DIR, "custom_models.json")
    if not os.path.exists(custom_models_path):
        return jsonify([])
    with open(custom_models_path, 'r', encoding='utf-8') as file:
        custom_models = json.load(file)
    custom_models = [model for model in custom_models if model["id"] != model_id]
    with open(custom_models_path, 'w', encoding='utf-8') as file:
        json.dump(custom_models, file)
    return jsonify(custom_models)